"""
"""


#[
from __future__ import annotations

from typing import (Self, Any, )
import numpy as _np
import prettytable as _pt

from .. import pages as _pages
from ..databoxes.main import (Databox, )
from ..dates import (Period, )
#]



class Mixin:
    """
    """
    #[

    @property
    def pretty(self, /, ) -> _pt.PrettyTable:
        """
        """
        return self.get_pretty()

    @property
    def pretty_full(self, /, ) -> str:
        """
        """
        return self.get_pretty(full=True, )

    def get_table(self, /, *args, **kwargs, ) -> _pt.PrettyTable:
        """
        """
        table = _pt.PrettyTable()
        table.field_names = self._TABLE_FIELDS
        table.align = "r"
        table.align["NAME"] = "l"
        for action in self._registers:
            register = self._get_register_by_name(action, )
            if register:
                self._add_register_to_table(
                    table, register, action,
                    *args, **kwargs,
                )
        return table

    get_pretty = get_table

    @_pages.reference(
        category="information",
    )
    def print_table(self, *args, **kwargs, ) -> None:
        r"""
................................................................................

==Print the `SimulationPlan` as a table==

    self.print_table()


### Input arguments ###

???+ input "self"
    `SimulationPlan` to be printed on the screen, with one row showing
    exogenized or endogenized data points grouped by the name and dates.


### Returns ###

Returns no value; the table is printed on the screen.

................................................................................
        """
        print(self.get_pretty(*args, **kwargs, ), )

    def pretty_print(self, *args, **kwargs, ) -> None:
        """
        """
        print(self.get_pretty(*args, **kwargs, ), )

    def get_pretty_string(self, *args, **kwargs, ) -> str:
        """
        """
        return self.get_pretty(*args, **kwargs, ).get_string()




"""
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Any, Callable, )
import dataclasses as _dc
import copy as _cp
import numpy as _np

from ..incidences import main as _incidence
from .. import equations as _equations
from .. import wrongdoings as _wrongdoings
from .. import makers as _makers

from . import _transforms as _transforms
#]


__all__ = (
    "Explanatory",
)


_RESIDUAL_NAME_FORMAT = "res_{lhs_name}"


@_dc.dataclass
class Explanatory:
    """
    """
    #[

    _IDENTITY_EQUAL_SIGN = "==="
    equation: _equations.Equation | None = None
    is_identity: bool | None = None
    _lhs_human: str | None = None
    lhs_name: str | None = None
    residual_name: str | None = None
    lhs_qid: int | None = None
    _rhs_human: str | None = None
    _lhs_transform: TransformProtocol | None = None
    _eval_level_str: str | None = None
    _eval_residual_str: str | None = None
    _context: dict[str, Callable] | None = None
    all_names: tuple(str) | None = None
    eval_level: Callable | None = None
    eval_residual: Callable | None = None

    def __init__(
        self,
        equation: _equations.Equation,
        /,
        context: dict[str, Callable] | None = None,
        residual_name_format: str | None = None,
        **kwargs,
    ) -> None:
        """
        """
        self._residual_name_format = (
            residual_name_format
            if residual_name_format
            else _RESIDUAL_NAME_FORMAT
        )
        self.equation = equation.copy()
        self._detect_identity()
        self._split_equation()
        self._parse_lhs()
        self._add_residual_to_rhs()
        self._collect_all_names()

    def finalize(
        self,
        name_to_qid,
        /,
    ) -> None:
        """
        """
        self.equation.finalize(name_to_qid, )
        self.lhs_qid = name_to_qid[self.lhs_name]
        self.residual_qid = name_to_qid[self.residual_name] if self.residual_name is not None else None
        rhs_xtring, *_ = _equations.xtring_from_human(self._rhs_human, name_to_qid, )
        lhs_xtring, *_ = _equations.xtring_from_human(self._lhs_human, name_to_qid, )
        lhs_token = _incidence.Token(self.lhs_qid, 0, )
        self._create_eval_level(lhs_token, rhs_xtring, )
        self._create_eval_residual(lhs_xtring, rhs_xtring, )

    @property
    def min_shift(self, /, ) -> int:
        """
        """
        return _incidence.get_min_shift(self.equation.incidence, )

    @property
    def max_shift(self, /, ) -> int:
        """
        """
        return _incidence.get_max_shift(self.equation.incidence, )


    @property
    def description(self, /, ) -> str:
        """
        """
        return (
            str(self.equation.description)
            if self.equation.description is not None else ""
        )

    def print_equation(
        self,
        /,
        indent: int = 0,
        description: bool = True,
    ) -> str:
        """
        """
        indented = " " * indent
        print_description = description and self.equation.description
        return (
            ((f"{indented}\"{self.equation.description}\"\n") if print_description else "")
            +f"{indented}{self.equation.human}"
        )

    def _detect_identity(self, /, ) -> None:
        """
        """
        self.is_identity = self._IDENTITY_EQUAL_SIGN in self.equation.human

    def _split_equation(self, ) -> None:
        """
        """
        human = self.equation.human
        equal_sign = "=" if not self.is_identity else self._IDENTITY_EQUAL_SIGN
        split = human.split(equal_sign, )
        if len(split) != 2:
            raise _wrongdoings.IrisPieError(
                f"This equation is not in a LHS=RHS or LHS{self._IDENTITY_EQUAL_SIGN}RHS form: {self.equation.human}"
            )
        self._lhs_human, self._rhs_human = split

    def _parse_lhs(self, ) -> None:
        lhs_transform, lhs_name = _transforms.recognize_transform_in_equation(self._lhs_human, )
        self._lhs_transform = lhs_transform
        self.lhs_name = lhs_name

    def _add_residual_to_rhs(self, /, ) -> None:
        """
        """
        if self.is_identity:
            self.residual_name = None
        else:
            self.residual_name = self._residual_name_format.format(lhs_name=self.lhs_name, )
            self._rhs_human += f"+{self.residual_name}"
            self.equation.human += f"+{self.residual_name}"

    def _collect_all_names(self, /, ) -> None:
        """
        """
        self.all_names = tuple(set(
            _equations.generate_names_from_human(self.equation.human, )
        )) + ((self.residual_name, ) if self.residual_name is not None else ())

    def _create_eval_level(
        self,
        lhs_token: str,
        rhs_xtring: str,
        /,
    ) -> None:
        args = ("x", "t", )
        body = self._lhs_transform.create_eval_level_str(lhs_token, rhs_xtring, )
        self.eval_level, self._eval_level_str, *_ = \
            _makers.make_lambda(args, body, self._context, )

    def _create_eval_residual(
        self,
        lhs_xtring: str,
        rhs_xtring: str,
        /,
    ) -> None:
        if self.is_identity:
            return
        args = ("x", "t", )
        body = f"{lhs_xtring}-({rhs_xtring})"
        self.eval_residual, self._eval_residual_str, *_ = \
            _makers.make_lambda(args, body, self._context, )

    def __str__(self, /, ) -> str:
        """
        """
        indented = " " * 4
        return "\n".join((
            f"{indented}Explanatory object",
            f"{indented}Description: \"{self.description}\"",
            f"{indented}Equation: {self.equation.human}",
            f"{indented}LHS name: {self.lhs_name}",
            f"{indented}LHS transform: {str(self._lhs_transform)}",
            f"{indented}Residual name: {self.residual_name}",
        ))

    def simulate(
        self,
        data: _np.ndarray,
        columns: int | _np.ndarray,
        values: int | _np.ndarray,
        /,
    ) -> dict[str, Any]:
        """
        """
        lhs_row = self.lhs_qid
        residual_row = self.residual_qid
        data[lhs_row, columns] = self.eval_level(data, columns, )
        is_finite = _np.isfinite(data[lhs_row, columns])
        return {
            "simulated_name": self.lhs_name,
            "is_finite": is_finite,
            "lhs_value": data[lhs_row, columns],
            "residual_value": data[residual_row, columns] if residual_row is not None else None,
        }

    def exogenize(
        self,
        data: _np.ndarray,
        columns: int | _np.ndarray,
        values: int | _np.ndarray,
        /,
    ) -> dict[str, Any]:
        """
        """
        lhs_row = self.lhs_qid
        residual_row = self.residual_qid
        data[lhs_row, columns] = values
        data[residual_row, columns] = self.eval_residual(data, columns, )
        is_finite = _np.isfinite(data[residual_row, columns])
        return {
            "simulated_name": self.residual_name,
            "is_finite": is_finite,
            "lhs_value": data[lhs_row, columns],
            "residual_value": data[residual_row, columns],
        }

    #]




"""
First-order system simulators
==============================

Table of contents
------------------

* `simulate`
* `simulate_flat`
* `simulate_conditional`
* `simulate_measurement`
* `_generate_period_system`
* `_store_smooth`
* `_get_init_xi`

"""


#[
from __future__ import annotations

from typing import (Protocol, )
import numpy as _np
import functools as _ft
import wlogging as _wl

from ..dates import (Period, )
from ..frames import (Frame, )
from ..dataslates.main import (Dataslate, )
from ..plans.simulation_plans import (SimulationPlan, )
from .. import frames as _frames
from ..incidences import main as _incidences
from .solutions import (Solution, )
from .descriptors import (SolutionVectors, Squid, )
from .kalmans import (Cache, )
from . import kalmans as _kalmans
from . import shock_simulators as _shock_simulators
#]


_extract_shock_values \
    = _shock_simulators.extract_shock_values


_simulate_anticipated_shocks \
    = _shock_simulators.simulate_square_anticipated_shocks


class _FordSimulatableProtocol(Protocol, ):
    """
    """
    #[

    num_variants: int

    solution_vectors: SolutionVectors

    def get_singleton_solution(self, deviation: bool, ) -> Solution: ...

    #]


def simulate(
    model_v: _FordSimulatableProtocol,
    dataslate_v: Dataslate,
    plan: SimulationPlan | None,
    vid: int,
    *,
    logger: _wl.Logger,
    #
    deviation: bool = False,
    check_singularity: bool = False,
    force_split_frames: bool = False,
    return_info: bool = False,
) -> dict[str, Any]:
    """
    """
    #[
    periods = dataslate_v.periods
    num_periods = dataslate_v.num_periods
    data_array = dataslate_v.get_data_variant(0, )
    squid = Squid(model_v, )
    #
    # In first-order simulation, splitting into frames is necessary only if
    # there are some endogenized anticipated shocks and, at the same time, some
    # nonzero unanticipated shocks (not counting the first period in either case)
    #
    is_plan_empty = plan is None or plan.is_empty
    needs_splitting = _check_needs_splitting(
        plan=plan,
        dataslate=dataslate_v,
        squid=squid,
        is_plan_empty=is_plan_empty,
        force_split_frames=force_split_frames,
    )
    logger.debug(f"Needs splitting: {needs_splitting}")
    #
    if needs_splitting:
        create_frames = _frames.split_into_frames
    else:
        create_frames = _frames.setup_single_frame
    #
    if is_plan_empty:
        simulate_frame = simulate_flat
    else:
        plan_registers = _get_plan_registers_as_bool_arrays(plan, periods, )
        simulate_frame = _ft.partial(
            simulate_conditional,
            check_singularity=check_singularity,
            input_data_array=data_array.copy(),
            plan_registers=plan_registers,
            Z_xi=_create_Z_xi(squid, )
        )
    #
    # Split the base span into frames
    # Each frame is simulated until the end of the dataslate base span
    #
    # TODO: Optimize the simulation end of each frame based on the presence of
    # endogenized anticipated shocks and nonzero anticipated shocks
    #
    dataslate_end = dataslate_v.base_periods[-1]
    frames = create_frames(
        model_v, dataslate_v, plan,
        get_simulation_end=lambda *_, : dataslate_end,
    )
    logger.debug(f"Splitting simulation into {len(frames)} frame(s)")
    #
    # Resolve columns within each frame based on the start date of the dataslate
    for frame in frames:
        frame.resolve_columns(periods[0], )
    #
    #
    #===========================================================================
    xi_array = _np.full((squid.num_xi, num_periods, ), _np.nan, )
    frame_db = ()
    for frame in frames:
        #
        if needs_splitting:
            frame_ds = dataslate_v.copy(invariant=False, )
            frame.reset_unanticipated(frame_ds, model_v, )
        else:
            frame_ds = dataslate_v
        #
        xi_array_in_frame = simulate_frame(
            model_v, frame_ds, frame, vid,
            logger=logger,
            deviation=deviation,
        )
        xi_array[:, frame.slice] = xi_array_in_frame[:, frame.slice]
        #
        simulate_measurement(
            model_v, dataslate_v, frame, xi_array, vid,
            logger=logger,
            deviation=deviation,
        )
        #
        # Copy frame data back to the main dataslate if there are more than one
        # frame; otherwise the main dataslate is the same as the frame dataslate
        if needs_splitting:
            frame.copy_frame_data(dataslate_v, frame_ds, )
        #
        # Create a frame databox only if output info is requested
        if return_info:
            frame_db = frame_db + (frame_ds.to_databox(), )
    #===========================================================================
    #
    #
    info = {
        "method": "first_order",
        "frames": frames,
        "frame_db": frame_db,
    }
    #
    return info
    #]


def simulate_flat(
    model_v: _FordSimulatableProtocol,
    dataslate_v: Dataslate,
    frame: Frame,
    vid: int,
    *,
    logger: _wl.Logger,
    deviation: bool,
) -> _np.ndarray:
    """
    """
    #[
    logger.debug(f"Running flat simulation {frame.start}>>{frame.end}>>{frame.simulation_end}")
    solution = model_v.get_singleton_solution(deviation=deviation, )
    vec = model_v.solution_vectors
    num_periods = dataslate_v.num_periods
    #
    dataslate_v.logarithmize()
    data_array = dataslate_v.get_data_variant(0, )
    #
    T = solution.T
    P = solution.P
    R = solution.R
    K = solution.K
    #
    Z = solution.Z
    H = solution.H
    D = solution.D
    #
    xi = _get_init_xi(data_array, frame.first, vec, )
    curr_xi_qids, curr_xi_indexes = vec.get_curr_transition_indexes()
    #
    # Prepare array of all xi values (needed for measurement simulation)
    num_xi = len(vec.transition_variables)
    xi_array = _np.full((num_xi, num_periods, ), _np.nan, )
    #
    u_array = _extract_shock_values(data_array, vec.unanticipated_shocks, )
    all_v_impact = _simulate_anticipated_shocks(model_v, dataslate_v, frame, )
    #
    # Simulate one period at a time
    # Store results in data_array
    # Capture xi_array for later use
    #
    for t in range(*frame.simulation_slice.indices(num_periods, ), ):
        xi = T @ xi + P @ u_array[:, t] + K
        if all_v_impact[t] is not None:
            xi += all_v_impact[t]
        xi_array[:, t] = xi
        data_array[curr_xi_qids, t] = xi[curr_xi_indexes, ...]
    #
    dataslate_v.delogarithmize()
    #
    return xi_array
    #]


def simulate_conditional(
    model_v: _FordSimulatableProtocol,
    dataslate_v: Dataslate,
    frame: Frame,
    vid: int,
    *,
    logger: _wl.Logger,
    deviation: bool,
    check_singularity: bool,
    #
    plan_registers: dict[str, _np.ndarray],
    input_data_array: _np.ndarray,
    Z_xi: _np.ndarray,
) -> _np.ndarray:
    """
    """
    #[
    logger.debug(f"Running conditional simulation {frame.start}>>{frame.end}>>{frame.simulation_end}")
    solution = model_v.get_singleton_solution(deviation=deviation, )
    vec = model_v.solution_vectors
    periods = dataslate_v.periods
    num_periods = dataslate_v.num_periods
    squid = Squid(model_v, )
    #
    dataslate_v.logarithmize()
    data_array = dataslate_v.get_data_variant()
    #
    # Prepare array of all xi values (needed for measurement simulation)
    xi_array = _np.full((squid.num_xi, num_periods, ), _np.nan, )
    #
    curr_xi_qids, curr_xi_indexes = vec.get_curr_transition_indexes()
    #
    # Initialize Kalman filter:
    # [initial median, initial MSE, number of initials to estimates, ]
    init_med = _get_init_xi(data_array, frame.first, vec, )
    init_mse = _np.zeros((squid.num_xi, squid.num_xi), )
    unknown_init_impact = None
    initials = (init_med, init_mse, unknown_init_impact, )
    #
    u0_array = data_array[squid.u_qids, frame.simulation_slice]
    v0_array = data_array[squid.v_qids, frame.simulation_slice]
    w0_array = data_array[squid.w_qids, frame.simulation_slice]
    #
    if model_v.is_deterministic:
        std_u_array = _np.ones((squid.num_u, frame.num_simulation_columns, ), )
        std_v_array = _np.ones((squid.num_v, frame.num_simulation_columns, ), )
        std_w_array = _np.ones((squid.num_w, frame.num_simulation_columns, ), )
    else:
        std_u_array = data_array[squid.std_u_qids, frame.simulation_slice]
        std_v_array = data_array[squid.std_v_qids, frame.simulation_slice]
        std_w_array = data_array[squid.std_w_qids, frame.simulation_slice]
    #
    all_v_impact = _simulate_anticipated_shocks(model_v, dataslate_v, frame, )
    #
    # Preallocate default data
    curr_xi_exogenized = _np.full((squid.num_curr_xi, num_periods, ), _np.nan, )
    std_u_endogenized = _np.zeros((squid.num_u, num_periods, ), )
    std_w_endogenized = _np.zeros((squid.num_w, num_periods, ), )
    #
    # Insert SimulationPlan information
    incidence_v, forward, Rx, std_v_endogenized = None, None, None, None
    if plan_registers:
        packed = (plan_registers, input_data_array, squid, frame.slice, logger, )
        _insert_exogenized_unanticipated(curr_xi_exogenized, packed, )
        _insert_exogenized_anticipated(curr_xi_exogenized, packed, )
        _insert_endogenized_unanticipated(std_u_endogenized, packed, )
        #
        incidence_v = plan_registers["endogenized_anticipated"][:, frame.simulation_slice]
        sum_incidence_v = incidence_v.sum()
        incidence_v = incidence_v if sum_incidence_v else None
        if incidence_v is not None:
            forward = incidence_v.any(axis=0, ).nonzero()[0].max()
            Rx = solution.expand_square_solution(forward, )
            std_v_endogenized = std_v_array[incidence_v, ]
        logger.debug(f"Endogenized anticipated: {sum_incidence_v}")
    #
    curr_xi_exogenized = curr_xi_exogenized[:, frame.simulation_slice]
    std_u_endogenized = std_u_endogenized[:, frame.simulation_slice]
    std_w_endogenized = std_w_endogenized[:, frame.simulation_slice]
    #
    # Define the Kalman callbacks
    generate_period_system = _ft.partial(
        _generate_period_system,
        initials=initials,
        solution=solution,
        Z_xi=Z_xi,
        curr_xi_exogenized=curr_xi_exogenized,
        std_u_endogenized=std_u_endogenized,
        std_w_endogenized=std_w_endogenized,
        std_v_endogenized=std_v_endogenized,
        all_v_impact=all_v_impact[frame.simulation_slice],
        incidence_v=incidence_v,
        Rx=Rx,
    )
    #
    generate_period_data = _ft.partial(
        _generate_period_data,
        curr_xi_exogenized=curr_xi_exogenized,
        u_array=u0_array,
        v_array=v0_array,
        w_array=w0_array,
    )
    store_smooth = _ft.partial(
        _store_smooth,
        data_array=data_array[:, frame.simulation_slice],
        xi_array=xi_array[:, frame.simulation_slice],
        incidence_v=incidence_v,
        squid=squid,
    )
    #
    # Run Kalman filter and smoother
    cache = _kalmans.predict(
        num_periods=frame.num_simulation_columns,
        generate_period_system=generate_period_system,
        generate_period_data=generate_period_data,
        store_smooth=True,
        check_singularity=check_singularity,
    )
    _kalmans.smooth(
        cache=cache,
        store_smooth=store_smooth,
    )
    dataslate_v.delogarithmize()
    #
    return xi_array
    #]


def simulate_measurement(
    model_v: _FordSimulatableProtocol,
    dataslate_v: Dataslate,
    frame: Frame,
    xi_array: _np.ndarray,
    vid: int,
    *,
    logger: _wl.Logger,
    deviation: bool = False,
    **kwargs,
) -> None:
    """
    """
    #[
    logger.debug("Running measurement simulation")
    solution = model_v.get_singleton_solution(deviation=deviation, )
    vec = model_v.solution_vectors
    num_periods = dataslate_v.num_periods
    #
    dataslate_v.logarithmize()
    data_array = dataslate_v.get_data_variant(0, )
    #
    Z = solution.Z
    H = solution.H
    D = solution.D if not deviation else 0
    #
    y_qids = [t.qid for t in vec.measurement_variables]
    w_array = _extract_shock_values(data_array, vec.measurement_shocks, )
    #
    for t in range(*frame.simulation_slice.indices(num_periods), ):
        data_array[y_qids, t] = Z @ xi_array[:, t] + H @ w_array[:, t] + D
    #
    dataslate_v.delogarithmize()
    #]


def _generate_period_system(
    t: int,
    #
    initials: list[_np.ndarray, _np.ndarray, int, ],
    solution: Solution,
    Z_xi: _np.ndarray | None,
    curr_xi_exogenized: _np.ndarray | None,
    std_u_endogenized: _np.ndarray | None,
    std_w_endogenized: _np.ndarray | None,
    std_v_endogenized: _np.ndarray | None,
    all_v_impact: Iterable[_np.ndarray | None],
    incidence_v: _np.ndarray | None,
    Rx: list[_np.ndarray] | None,
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    #
    # Initial conditions
    if t == -1:
        return _generate_initials(initials, std_v_endogenized, )
    #
    # Period t system matrices
    T = solution.T
    P = solution.P
    K = solution.K
    Z = _generate_Z(t, Z_xi, curr_xi_exogenized, )
    num_xi = T.shape[0]
    num_y = Z.shape[0]
    H = _np.zeros((num_y, solution.num_w, ), )
    D = _np.zeros((num_y, ), )
    cov_u = _np.diag(std_u_endogenized[:, t]**2, )
    cov_w = _np.diag(std_w_endogenized[:, t]**2, )
    v_impact = all_v_impact[t]
    #
    if incidence_v is not None and incidence_v.any():
        #
        # Endogenized anticipated shocks
        R = _generate_R(t, incidence_v, Rx, num_xi, )
        num_v_endogenized = R.shape[1]
        T = _np.block([
            [T, R],
            [_np.zeros((num_v_endogenized, num_xi)), _np.eye(num_v_endogenized)],
        ])
        P = _np.pad(P, ((0, num_v_endogenized), (0, 0)), )
        K = _np.pad(K, (0, num_v_endogenized), )
        Z = _np.pad(Z, ((0, 0), (0, num_v_endogenized)), )
        if v_impact is not None:
            v_impact = _np.pad(v_impact, (0, num_v_endogenized), )
    #
    return T, P, K, Z, H, D, cov_u, cov_w, v_impact,
    #]


def _generate_initials(
    initials: list[_np.ndarray, _np.ndarray, int, ],
    std_v_endogenized: _np.ndarray,
) -> tuple[_np.ndarray, _np.ndarray]:
    """
    """
    #[
    num_v_endogenized = (
        len(std_v_endogenized)
        if std_v_endogenized is not None else 0
    )
    #
    # Extend initial conditions to include endogenized anticipated shocks
    if num_v_endogenized > 0:
        init_med = initials[0]
        init_mse = initials[1]
        uknown_init_impact = None
        initials = (
            _np.pad(initials[0], (0, num_v_endogenized), ),
            _block_diag(initials[1], _np.diag(std_v_endogenized**2), ),
            uknown_init_impact,
        )
    #
    # Return updated initial conditions
    return initials
    #]


def _generate_Z(
    t: int,
    Z_xi: _np.ndarray | None,
    curr_xi_exogenized: _np.ndarray | None,
) -> _np.ndarray:
    """
    """
    #[
    if curr_xi_exogenized is not None:
        inx_y = ~_np.isnan(curr_xi_exogenized[:, t])
        return Z_xi[inx_y, :]
    else:
        return _np.zeros((0, Z_xi.shape[1], ), )
    #]


def _generate_R(
    t: int,
    incidence_v: _np.ndarray,
    Rx: list[_np.ndarray],
    num_xi: int,
) -> _np.ndarray:
    """
    """
    #[
    def zeros(i):
        return _np.zeros((num_xi, i.sum()), )
    Rx_t = [None]*t + Rx[:len(Rx)-t]
    return _np.hstack(tuple(
        r[:, i] if r is not None else zeros(i)
        for r, i in zip(Rx_t, incidence_v.T, )
    ))
    #]


def _generate_period_data(
    t: int,

    curr_xi_exogenized: _np.ndarray | None,
    u_array: _np.ndarray,
    v_array: _np.ndarray,
    w_array: _np.ndarray,
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    if curr_xi_exogenized is not None:
        inx_y = ~_np.isnan(curr_xi_exogenized[:, t], )
        y = curr_xi_exogenized[inx_y, t]
    else:
        y = _np.zeros((0, ), )
    u = u_array[:, t]
    v = v_array[:, t]
    w = w_array[:, t]
    return y, u, v, w,
    #]


def _store_smooth(
    data_array: _np.ndarray,
    xi_array: _np.ndarray,
    incidence_v: _np.ndarray | None,
    squid: Squid,
    #
    t: int,
    xi: _np.ndarray | None = None,
    u: _np.ndarray | None = None,
    v: _np.ndarray | None = None,
    w: _np.ndarray | None = None,
    **kwargs,
) -> None:
    """
    """
    #[
    def update_v_endogenized(v_endogenized, ):
        ___ = data_array[squid.v_qids, :]
        ___[incidence_v] += v_endogenized
        data_array[squid.v_qids, :] = ___
    #
    if xi is not None:
        if incidence_v is not None:
            num_v_endogenized = incidence_v.sum()
            v_endogenized = xi[-num_v_endogenized:]
            xi = xi[:-num_v_endogenized]
            if t == data_array.shape[1] - 1:
                update_v_endogenized(v_endogenized, )
        data_array[squid.curr_xi_qids, t] = xi[squid.curr_xi_indexes, ...]
        xi_array[:, t] = xi
    if u is not None:
        data_array[squid.u_qids, t] = u
    if w is not None:
        data_array[squid.w_qids, t] = w
    #]


def _get_init_xi(
    maybelog_working_data: _np.ndarray,
    first_column: int,
    solution_vectors: SolutionVectors,
    /,
) -> _np.ndarray:
    """
    """
    #[
    init_xi_rows, init_xi_columns \
        = _incidences.rows_and_columns_from_tokens(
            solution_vectors.transition_variables,
            first_column - 1,
        )
    init_xi = maybelog_working_data[init_xi_rows, init_xi_columns]
    unnecessary_initials = ~_np.array(solution_vectors.are_initial_conditions, )
    init_xi[unnecessary_initials] = 0
    return init_xi
    #]


def _get_plan_registers_as_bool_arrays(
    plan: SimulationPlan,
    periods: tuple[Period, ...],
) -> dict[str, _np.ndarray]:
    """
    """
    #[
    register_names = (
        "exogenized_anticipated",
        "endogenized_anticipated",
        "exogenized_unanticipated",
        "endogenized_unanticipated",
    )
    get_register = _ft.partial(
        plan.get_register_as_bool_array,
        names=...,
        periods=periods,
    )
    return {
        n: get_register(register_name=n, )
        for n in register_names
    }
    #]


def _create_Z_xi(squid, /, ) -> _np.ndarray:
    """
    """
    #[
    Z_xi = _np.zeros((squid.num_curr_xi, squid.num_xi, ), )
    Z_xi[range(squid.num_curr_xi, ), squid.curr_xi_indexes] = 1
    return Z_xi
    #]


def _insert_exogenized_unanticipated(
    curr_xi_exogenized: _np.ndarray,
    packed: tuple,
) -> None:
    """
    """
    #[
    plan_registers, input_data_array, squid, frame_slice, logger = packed
    #
    ___ = plan_registers["exogenized_unanticipated"][:, frame_slice]
    curr_xi_exogenized[:, frame_slice][___] \
        = input_data_array[squid.curr_xi_qids, frame_slice][___]
    #
    logger.debug(f"Exogenized unanticipated: {___.sum()}")
    #]


def _insert_exogenized_anticipated(
    curr_xi_exogenized: _np.ndarray,
    packed: tuple,
) -> None:
    """
    """
    #[
    plan_registers, input_data_array, squid, frame_slice, logger = packed
    #
    ___ = plan_registers["exogenized_anticipated"][:, frame_slice]
    curr_xi_exogenized[:, frame_slice][___] \
        = input_data_array[squid.curr_xi_qids, frame_slice][___]
    #
    logger.debug(f"Exogenized anticipated: {___.sum()}")
    #]


def _insert_endogenized_unanticipated(
    std_u_endogenized: _np.ndarray,
    packed: tuple,
) -> None:
    """
    """
    #[
    plan_registers, input_data_array, squid, frame_slice, logger = packed
    #
    ___ = plan_registers["endogenized_unanticipated"][:, frame_slice]
    std_u_endogenized[:, frame_slice][___] \
        = input_data_array[squid.std_u_qids, frame_slice][___]
    #
    logger.debug(f"Endogenized unanticipated: {___.sum()}")
    #]



def _block_diag(A, B):
    """
    Concatenate two square matrices into a block diagonal matrix
    """
    #[
    return _np.block([
        [A, _np.zeros((A.shape[0], B.shape[1]), )],
        [_np.zeros((B.shape[0], A.shape[1]), ), B],
    ])
    #]


def _check_needs_splitting(
    plan: SimulationPlan | None,
    dataslate: Dataslate,
    squid: Squid,
    is_plan_empty: bool,
    force_split_frames: bool,
) -> bool:
    """
    """
    #[
    if force_split_frames:
        return True
    #
    if is_plan_empty and not force_split_frames:
        return False
    #
    any_endogenized_anticipated = (
        plan is not None
        and plan.any_endogenized_anticipated_except_start
    )
    if not any_endogenized_anticipated:
        return False
    #
    base_columns_except_start = dataslate.base_columns[1:]
    data_array = dataslate.get_data_variant(0, )
    any_nonzero_unanticipated_shocks \
            = data_array \
            [squid.v_qids, :] \
            [:, base_columns_except_start] \
            .any()
    if not any_nonzero_unanticipated_shocks:
        return False
    #
    return True
    #]




"""
Time series chartpacks
"""


#[
from __future__ import annotations

from typing import (Self, Iterator, Iterable, Sequence, Callable, )
from types import (EllipsisType, )
import functools as _ft
import re as _re
import math as _ma
import plotly.graph_objects as _pg

from ..conveniences import descriptions as _descriptions
from ..conveniences import copies as _copies
from .. import plotly_wrap as _plotly_wrap
from .. import dates as _dates
from .. import pages as _pages
from ..databoxes import main as _databoxes
#]


__all__ =  (
    "Chartpack",
)


_FIGURE_SETTINGS = {
    "legend": None,
    "highlight": None,
    "tiles": None,
    "shared_xaxes": False,
}
_FIGURE_SETTINGS_KEYS = tuple(_FIGURE_SETTINGS.keys(), )


_CHART_SETTINGS = {
    "transforms": {},
    "reverse_plot_order": False,
    "span": ...,
    "chart_type": "line",
}
_CHART_SETTINGS_KEYS = tuple(_CHART_SETTINGS.keys(), )


_CHART_INPUT_STRING_PATTERN = _re.compile(
    r"^(?:(?P<title>.*?):)?(?P<expression>.*?)(?:\[(?P<transform>.*?)\])?$"
)


@_pages.reference(
    path=("visualization_reporting", "chartpacks.md", ),
    categories={
        "constructor": "Creating new chartpacks",
        "property": None,
        "plot": "Plotting chartpacks",
        "add": "Adding figures and charts to chartpacks",
    }
)
class Chartpack(
    _descriptions.DescriptionMixin,
    _copies.Mixin,
):
    """
················································································

Chartpacks
===========

················································································
    """
    #[

    __slots__ = (
        "title",
        "_figures",
        "_description",
        "_figure_settings",
        "_chart_settings",
    )

    def __init__(
        self,
        title: str = "",
        **kwargs,
    ) -> None:
        """
        """
        self.title = title
        self._figures = None
        self._description = None
        #
        self._figure_settings = {
            n: kwargs[n]
            for n in _FIGURE_SETTINGS_KEYS
            if n in kwargs
        }
        #
        self._chart_settings = {
            n: kwargs[n]
            for n in _CHART_SETTINGS_KEYS
            if n in kwargs
        }

    @classmethod
    @_pages.reference(category="constructor", call_name="Chartpack", )
    def _constructor_doc():
        """
················································································

==Create a new chartpack==

```
self = Chartpack(
    title="",
    span=...,
    tiles=None,
    transforms=None,
    highlight=None,
    legend=None,
    reverse_plot_order=False,
)
```

### Input arguments ###


???+ input "title"
    The title of the chartpack, used as a basis for creating a caption
    shown at the top of each figure.

???+ input "span"
    The date span on which the time series will be plotted.

???+ input "tiles"
    The number of rows and columns of the figure grid. If input "None", the number of
    rows and columns will be determined automatically.

???+ input "transforms"
    A dictionary of functions that will be applied to the input data before
    plotting.

???+ input "highlight"
    A date span that will be highlighted in the charts.

???+ input "legend"
    A list of strings that will be used as the legend for the charts.

???+ input "reverse_plot_order"
    If `True`, the order of plotting the individual time series within each
    chart will be reversed.


### Returns


???+ returns "self"

    A new empty `Chartpack` object.

················································································
        """
        pass

    @_pages.reference(category="plot", )
    def plot(
        self,
        input_db: _databoxes.Databox,
        show_figures: bool = True,
        **kwargs,
    ) -> tuple[_pg.Figure, ...]:
        """
················································································

==Plot the chartpack==

················································································
        """
        figures = tuple(
            figure.plot(input_db, **self.__dict__, )
            for figure in self
        )
        if show_figures:
            for f in figures: f.show()
        return figures

    def format_figure_titles(
        self,
        **kwargs,
    ) -> None:
        """
        """
        for figure in self:
            figure.format_figure_title(**kwargs, )

    @property
    @_pages.reference(category="property", )
    def num_figures(self, ) -> int:
        """==Total number of figures in the chartpack=="""
        return len(self._figures, ) if self._figures else 0

    def _add_figure(self, figure: _Figure, ) -> None:
        """
        """
        if not self._figures:
            self._figures = []
        self._figures.append(figure, )

    @_pages.reference(category="add", )
    def add_figure(self, figure_string: str, **kwargs, ) -> None:
        """
················································································

==Add a new figure to the chartpack==

················································································
        """
        new_figure = _Figure.from_string(
            figure_string,
            kwargs=kwargs,
            figure_settings_cascaded=self._figure_settings,
            chart_settings_cascaded=self._chart_settings,
        )
        self._add_figure(new_figure, )
        return new_figure

    @staticmethod
    @_pages.reference(category="add", call_name="add_chart", )
    def _add_chart_doc():
        """
················································································

==Add a new chart to an existing figure in the chartpack==

················································································
        """
        pass

    def __str__(self, /, ) -> str:
        """
        """
        return repr(self, )

    def __repr__(self, /, ) -> str:
        """
        """
        return _tree_repr(self, )

    def _one_liner(self, /, ) -> str:
        return f"Chartpack({self.title!r}, )"

    def __getitem__(self, item: str | int, ) -> _Figure:
        """
        """
        if isinstance(item, int, ):
            return self._figures[item]
        if isinstance(item, str, ):
            return next(f for f in self._figures if f.title == item)

    def __iter__(self, ) -> Iterator[_Figure]:
        """
        """
        return iter(self._figures, )

    #]


class _Figure:
    """
    """
    #[

    __slots__ = (
        "title",
        "_charts",
        "_chart_settings",
    ) + _FIGURE_SETTINGS_KEYS

    def __init__(
        self,
        title: str | None = None,
    ) -> None:
        """
        """
        for n in self.__slots__:
            setattr(self, n, None, )
        self.title = title
        self._charts = []
        self._chart_settings = {}

    @classmethod
    def from_string(
        klass,
        input_string: str,
        /,
        kwargs: dict[str, Any],
        figure_settings_cascaded: dict[str, Any],
        chart_settings_cascaded: dict[str, Any],
    ) -> Self:
        """
        """
        self = klass(title=input_string, )
        #
        for key in _FIGURE_SETTINGS_KEYS:
            if key in kwargs:
                setattr(self, key, kwargs[key], )
                continue
            if key in figure_settings_cascaded:
                setattr(self, key, figure_settings_cascaded[key], )
                continue
            else:
                setattr(self, key, _FIGURE_SETTINGS[key], )
                continue
        #
        for key in _CHART_SETTINGS_KEYS:
            if key in kwargs:
                self._chart_settings[key] = kwargs[key]
                continue
            if key in chart_settings_cascaded:
                self._chart_settings[key] = chart_settings_cascaded[key]
                continue
        #
        return self

    @property
    def num_charts(self, ) -> int:
        return len(self._charts, )

    def plot(
        self,
        input_db: _databoxes.Databox,
        /,
        shared_xaxes: bool = False,
        **kwargs,
    ) -> None:
        """
        """
        tiles = _resolve_tiles(self.tiles, self.num_charts, )
        figure = _plotly_wrap.make_subplots(
            rows=tiles[0],
            columns=tiles[1],
            subplot_titles=tuple(chart.caption for chart in self),
            shared_xaxes=self.shared_xaxes,
        )
        for i, chart in enumerate(self, ):
            chart.plot(
                input_db, figure, i,
                legend=self.legend if i == 0 else None,
            )
            if self.highlight is not None:
                _plotly_wrap.highlight(figure, self.highlight, subplot=i, )
        figure.update_layout(title={"text": self.title, }, )
        return figure

    def format_figure_title(
        self,
        **kwargs,
    ) -> None:
        """
        """
        self.title = self.title.format(**kwargs, )

    def __str__(self, /, ) -> str:
        """
        """
        return repr(self, )

    def __repr__(self, /, ) -> str:
        """
        """
        return _tree_repr(self, )

    def _one_liner(self, /, ) -> str:
        return f"Figure({self.title!r}, )"

    def _add_chart(self, chart: _Chart, ) -> None:
        """
        """
        self._charts.append(chart, )

    def add_charts(self, chart_strings: Iterable[str], **kwargs, ) -> None:
        """
        """
        for chart_string in chart_strings:
            self.add_chart(chart_string, **kwargs, )

    def add_chart(self, chart_string: str, **kwargs, ) -> None:
        """
        """
        chart = _Chart.from_string(
            chart_string,
            kwargs=kwargs,
            chart_settings_cascaded=self._chart_settings,
        )
        self._add_chart(chart, )

    def __iter__(self, /, ) -> Iterator[_Chart]:
        """
        """
        return iter(self._charts, )

    def __enter__(self, /, ) -> Self:
        return self

    def __exit__(self, *args, **kwargs, ) -> None:
        pass

    #]


class _Chart:
    """
    """
    #[

    __slots__ = (
        "title",
        "expression",
        "transform"
    ) + _CHART_SETTINGS_KEYS

    def __init__(
        self,
        expression: str,
        title: str | None = None,
        transform: str | Callable | None = None,
    ) -> None:
        """
        """
        for n in self.__slots__:
            setattr(self, n, None, )
        self.expression = (expression or "").strip()
        self.title = (title or "").strip() or self.expression
        self.transform = (transform or "").strip()

    @classmethod
    def from_string(
        klass,
        input_string: str,
        kwargs: dict[str, Any],
        chart_settings_cascaded: dict[str, Any],
    ) -> Self:
        """
        """
        match = _CHART_INPUT_STRING_PATTERN.match(input_string, )
        if not match:
            raise ValueError(f"Invalid input string: {input_string!r}")
        self = klass(**match.groupdict(), )
        #
        for key in _CHART_SETTINGS_KEYS:
            if key in kwargs:
                setattr(self, key, kwargs[key], )
                continue
            elif key in chart_settings_cascaded:
                setattr(self, key, chart_settings_cascaded[key], )
                continue
            else:
                setattr(self, key, _CHART_SETTINGS[key], )
                continue
        #
        return self

    @property
    def caption(self, ) -> str:
        """Caption shown at the top of the chart"""
        return self.title or (self.expression + f" [{self.transform}]" if self.transform else "")

    def plot(
        self,
        input_db: _databoxes.Databox,
        figure: _pg.Figure,
        index: int,
        legend: Iterable[str, ...] | None = None,
    ) -> None:
        """
        """
        series_to_plot = input_db.evaluate_expression(self.expression, )
        series_to_plot = self._apply_transform(series_to_plot, )
        series_to_plot.plot(
            figure=figure,
            subplot=index,
            span=self.span,
            show_figure=False,
            freeze_span=True,
            legend=legend,
            reverse_plot_order=self.reverse_plot_order,
            chart_type=self.chart_type,
        )

    def _apply_transform(self, x, ):
        """
        """
        if self.transform:
            func = self.transforms[self.transform]
        else:
            func = None
        return func(x) if func else x

    def __str__(self, /, ) -> str:
        """
        """
        return repr(self, )

    def __repr__(self, /, ) -> str:
        """
        """
        return _tree_repr(self, )

    def _one_liner(self, /, ) -> str:
        return f"Chart({self.title!r}, {self.expression!r}, {self.transform!r}, )"

    def __iter__(self, ) -> NoReturn:
        yield from ()

    #]


def _add_strings(
    self,
    input_strings: Iterable[str] | str,
    call: Callable,
    **kwargs,
) -> None:
    """
    """
    #[
    if isinstance(input_strings, str, ):
        input_strings = (input_strings, )
    for s in input_strings:
        call(s, )
    #]


def _resolve_tiles(tiles, num_charts, ) -> tuple[int, int]:
    """
    """
    #[
    if tiles is None:
        return _auto_tiles(num_charts, )
    if isinstance(tiles, int, ):
        return _auto_tiles(tiles, )
    if isinstance(tiles, Sequence, ):
        return tiles[:2]
    raise TypeError(f"Invalid tiles: {tiles!r}")
    #]


def _auto_tiles(num_charts, ) -> tuple[int, int]:
    """
    """
    #[
    n = _ma.ceil(_ma.sqrt(num_charts, ), )
    if n * (n-1) >= num_charts:
        return (n, n-1, )
    return (n, n, )
    #]


_TREE_INDENT = "    |"
_TREE_BULLET = "--"


def _tree_repr(self, /, indent: tuple[str] = (), is_last: bool = False, ) -> str:
    """
    """
    #[
    child_indent = _create_child_indent(indent, is_last, )
    child_repr_func = _ft.partial(_tree_repr, indent=child_indent, )
    child_objects = tuple(c for c in self)
    num_child_objects = len(child_objects)
    child_tree_repr = tuple(
        _tree_repr(c, indent=child_indent, is_last=(i+1 == len(child_objects)), )
        for i, c in enumerate(child_objects)
    )
    indent_str = "".join(indent, )
    header = indent_str + (_TREE_BULLET if indent_str else "") + self._one_liner()
    return "\n".join((header, *child_tree_repr))
    #]


def _create_child_indent(indent: tuple[str], is_last: bool, ) -> str:
    """
    """
    #[
    if is_last and indent:
        indent = indent[:-1] + (" "*len(_TREE_INDENT), )
    return indent + (_TREE_INDENT, )
    #]




"""
Time frames for dynamic simulations
"""


#[
from __future__ import annotations

from typing import (Callable, Iterable, )
import numpy as _np

from .dates import (Period, )
from .dataslates.main import (Dataslate, )
from .plans.simulation_plans import (SimulationPlan, )
#]


class Frame:
    """
    """
    #[

    __slots__ = (
        # Periods
        "start",
        "end",
        "simulation_end",
        # Columns
        "first",
        "last",
        "simulation_last",
        "slice",
        "simulation_slice",
        "num_simulation_columns",
    )

    def __init__(
        self,
        start: Period,
        end: Period,
        simulation_end: Period | None = None,
    ) -> None:
        """
        """
        self.start = start
        self.end = end
        self.simulation_end = simulation_end
        self.first = None
        self.last = None
        self.simulation_last = None
        self.slice = None
        self.simulation_slice = None
        self.num_simulation_columns = None

    def __repr__(self, ) -> str:
        """
        """
        return f"Frame({self.start}, {self.end}, {self.simulation_end})"

    def resolve_columns(
        self,
        first_column_period: Period,
    ) -> None:
        """
        Given the time period of the first column of data array, determine the
        first and last columns of the frame
        """
        #[
        self.first = self.start - first_column_period
        self.last = self.end - first_column_period
        self.simulation_last = self.simulation_end - first_column_period
        self.slice = slice(self.first, self.last+1, )
        self.simulation_slice = slice(self.first, self.simulation_last+1, )
        self.num_simulation_columns = self.simulation_last - self.first + 1
        #]

    def reset_unanticipated(
        self,
        ds: Dataslate,
        model: Slatable,
    ) -> None:
        """
        """
        #[
        if self.start == self.simulation_end:
            return
        plannable = model.get_simulation_plannable()
        name_to_qid = model.create_name_to_qid()
        if not hasattr(plannable, "can_be_endogenized_unanticipated"):
            return
        unanticipated_names = plannable.can_be_endogenized_unanticipated
        unanticipated_qids = tuple(name_to_qid[name] for name in unanticipated_names)
        data = ds.get_data_variant(0, )
        data[unanticipated_qids, self.first+1:] = 0

    def copy_frame_data(
        self,
        main_ds: Dataslate,
        frame_ds: Dataslate,
    ) -> None:
        """
        Copy frame columns from frame_ds to main_ds
        """
        main_data = main_ds.get_data_variant(0, )
        frame_data = frame_ds.get_data_variant(0, )
        main_data[:, self.slice] = frame_data[:, self.slice]
    #]


def split_into_frames(
    model,
    dataslate: Dataslate,
    plan: SimulationPlan | None,
    get_simulation_end: Callable,
) -> tuple[Frame, ...]:
    """
    """
    if not model.is_singleton:
        raise ValueError("Model must be a singleton")
    if not dataslate.is_singleton:
        raise ValueError("Dataslate must be a singleton")
    base_periods = dataslate.base_periods
    break_points = _get_break_points_in_base_columns(model, dataslate, plan, )
    break_periods = _get_break_periods_from_break_points(break_points, base_periods, )
    break_periods_next = break_periods[1:] + (base_periods[-1] + 1, )
    return tuple(
        Frame(start, next_-1, get_simulation_end(start, next_-1, ), )
        for start, next_ in zip(break_periods, break_periods_next, )
    )


def setup_single_frame(
    model,
    dataslate: Dataslate,
    plan: SimulationPlan | None,
    get_simulation_end: Callable,
) -> tuple[Frame, ...]:
    """
    Return a tuple with a single frame that spans the entire simulation period
    """
    base_periods = dataslate.base_periods
    frame_start = base_periods[0]
    frame_end = base_periods[-1]
    simulation_end = frame_end
    single_frame = Frame(
        frame_start,
        frame_end,
        get_simulation_end(frame_start, frame_end, ),
    )
    return (single_frame, )


def _get_break_points_in_base_columns(
    model,
    dataslate: Dataslate | None,
    plan: SimulationPlan | None,
) -> tuple[Frame]:
    """
    """
    plannable = model.get_simulation_plannable()
    can_be_endogenized = plannable.can_be_endogenized_unanticipated
    can_be_exogenized = plannable.can_be_exogenized_unanticipated
    name_to_row = dataslate.create_name_to_row()
    rows = tuple(name_to_row[name] for name in can_be_endogenized)
    periods = dataslate.base_periods
    columns = dataslate.base_columns
    break_points = _np.zeros((len(columns), ), dtype=bool, )
    break_points[0] = True
    if rows and columns:
        cutout = dataslate.get_data_variant(0, )[rows, :][:, columns]
        break_points = _update_break_points(break_points, cutout, )
    #
    if plan is not None:
        cutout, *_ = plan.get_register_as_bool_array(
            "endogenized_unanticipated", can_be_endogenized, periods,
        )
        break_points = _update_break_points(break_points, cutout, )
        # cutout, *_ = plan.get_register_as_bool_array(
        #     "exogenized_unanticipated", can_be_exogenized, periods,
        # )
        # break_points = _update_break_points(break_points, cutout, )
        #
    return break_points


def _get_break_periods_from_break_points(
    break_points: Iterable[bool],
    base_periods: Iterable[Period],
    /,
) -> tuple[Period]:
    after_period = base_periods[-1] + 1
    zipped = zip(base_periods, break_points)
    return tuple(t for t, flag in zipped if flag)


def _update_break_points(break_points, new_array, ):
    """
    """
    if new_array is None:
        return break_points
    new_break_points = _np.any(_np.isfinite(new_array) & (new_array != 0), axis=0, )
    return break_points | new_break_points




"""
"""


#[
from __future__ import annotations

from typing import (Callable, )
from numbers import (Real, )
import dataclasses as _dc
import copy as _cp
import numpy as _np

from ..dates import (Period, Frequency, )
from ..conveniences import descriptions as _descriptions
from ..series import main as _series
from ..databoxes import main as _databoxes
from .. import has_variants as _has_variants
#]


class Invariant:
    """
    """
    #[

    __slots__ = (
        "names",
        "periods",
        "descriptions",
        "logly_indexes",
        "base_columns",
        "output_qids",
        "min_max_shift",
    )

    def __init__(
        self,
        names: Iterable[str],
        periods: Iterable[Period],
        /,
        *,
        base_columns: Iterable[int] | None = None,
        descriptions: Iterable[str | None] | None = None,
        output_names: Iterable[str] | None = None,
        qid_to_logly: dict[int, bool | None] | None = None,
        min_max_shift: tuple[int, int] = (0, 0),
        frequency: Frequency | None = None,
    ) -> None:
        """
        """
        self.names = tuple(names)
        self.periods = tuple(periods)
        self.base_columns = tuple(sorted(base_columns)) if base_columns else ()
        self._populate_descriptions(descriptions, )
        self._populate_logly_index(qid_to_logly, )
        self.min_max_shift = min_max_shift
        self._populate_output_qids(output_names, )

    def _populate_output_qids(
        self,
        output_names: Iterable[str] | None,
    ) -> None:
        """
        """
        output_names = (
            tuple(set(output_names) & set(self.names))
            if output_names is not None
            else self.names
        )
        self.output_qids = tuple(
            qid
            for qid, name in enumerate(self.names, )
            if name in output_names
        )

    @property
    def num_periods(self, /, ) -> int:
        """
        """
        return len(self.periods)

    @property
    def num_names(self, /, ) -> int:
        """
        """
        return len(self.names)

    @property
    def from_until(self, /, ) -> tuple[Period, Period]:
        """
        """
        return self.periods[0], self.periods[-1]

    from_to = from_until

    @property
    def base_periods(self, /, ) -> tuple[Period]:
        """
        """
        return tuple(self.periods[i] for i in self.base_columns)

    @base_periods.setter
    def base_periods(self, base_periods: Iterable[Period], /, ) -> None:
        """
        """
        base_periods = set(base_periods)
        self.base_columns = tuple(
            i for i, period in enumerate(self.periods)
            if period in base_periods
        )

    @property
    def nonbase_columns(self, /, ) -> tuple[int]:
        """
        """
        all_columns = set(range(self.num_periods))
        return tuple(sorted(all_columns - set(self.base_columns)))

    @property
    def base_slice(self, /, ) -> slice:
        """
        """
        base_columns = self.base_columns
        return slice(base_columns[0], base_columns[-1]+1)

    def copy(self, /, ) -> Invariant:
        """
        """
        return _cp.deepcopy(self, )

    def create_name_to_row(
        self,
        /,
    ) -> dict[str, int]:
        return {
            name: row
            for row, name in enumerate(self.names, )
        }

    def _populate_descriptions(
        self,
        descriptions: Iterable[str | None] | None,
    ) -> None:
        """
        """
        if descriptions is None:
            self.descriptions = tuple("" for _ in self.names)
        else:
            self.descriptions = tuple((d or "") for d in descriptions)

    def _populate_logly_index(
        self,
        qid_to_logly: dict[int, bool] | None,
        /,
    ) -> None:
        """
        """
        qid_to_logly = qid_to_logly or {}
        self.logly_indexes = tuple(
            i for i in range(len(self.names), )
            if qid_to_logly.get(i, False)
        )

    def remove_periods_from_start(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove:
            self.periods = self.periods[remove:]
            self.base_columns = tuple(
                i - remove for i in self.base_columns
                if i >= remove
            )

    def remove_periods_from_end(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove:
            self.periods = self.periods[:-remove]
            num_periods = len(self.periods)
            self.base_columns = tuple(
                i for i in self.base_columns
                if i < num_periods
            )





"""
Autoregressive extrapolation
"""


#[
from __future__ import annotations

from typing import (TYPE_CHECKING, )
from numbers import (Number, )
import numpy as _np
import scipy as _sp

from .. import pages as _pages
from .. import dates as _dates
from . import _functionalize

if TYPE_CHECKING:
    from typing import (Iterable, )
    from numbers import (Real, )
    from ..dates import (Period, )
#]


__all__ = ()


class Inlay:
    """
    """
    #[

    @_pages.reference(category="homogenizing", )
    def extrapolate(
        self: Self,
        ar_coeffs: Iterable[Real],
        span: Iterable[Period],
        *,
        intercept: Real = 0,
        log: bool = False,
    ) -> None:
        r"""
......................................................................

==Extrapolate time series using autoregressive process==


### Functional forms creating a new time `Series` object ###


    new = extrapolate(
        self,
        ar_coeffs,
        span,
        *,
        intercept=0,
        log=False,
    )


### Class methods changing an existing time `Series` object in-place ###


    self.extrapolate(
        ar_coeffs,
        span,
        *,
        intercept=0,
        log=False,
    )


### Input arguments ###


???+ input "self"
    The time `Series` object to be extrapolated by an autoregressive process.

???+ input "ar_coeffs"
    The autoregressive coefficients to be used in the extrapolation, entered as
    a tuple of AR_1, AR_2, ..., AR_p coefficients as if on the RHS of the AR
    process definition; see Details below.

???+ input "span"
    The time span on which the time series will be extrapolated.

???+ input "intercept"
    The intercept in the autorergressive process.

???+ input "log"
    If `log=True`, the time series will be logarithmized before the
    extrapolation and then delogarithmized back.


### Returns ###


???+ returns "self"
    The existing time `Series` object with its values replaced in-place.

???+ returns "new"
    A new time `Series` object.


### Details ###


The new extrapolated observations are created using this $p$-th order
autoregressive process defined recursively as:

$$
x_t = \rho_1 \, x_{t-1} + \cdots + \rho_p \, x_{t-p} + c, \qquad t = 1, \dots, T
$$

where

* the initial condion $x_{0},\ x_{-1}, \, \dots,\ x_{-p+1}$ are taken from
the existing observations in the input series `self`;

* the autoregressive coefficents
$\rho_1,\ \rho_2,\ \dots,\ \rho_p$ given by the input argument `ar_coeff`

* $c$ is the `intercept`.

......................................................................
        """
        if self.start is None:
            return
        span = self._resolve_dates(span, )
        if not span:
            return
        num_periods = len(span)
        ar_coeffs = (ar_coeffs, ) if isinstance(ar_coeffs, Number) else ar_coeffs
        order = len(ar_coeffs)
        ar = (1, ) + tuple(-i for i in ar_coeffs)
        start = span[0]
        initial_from_until = (start-order, start-1, )
        iter_initial_data = self.iter_own_data_variants_from_until(initial_from_until, )
        new_data = _np.hstack(tuple(
            _extrapolate(
                initial_data, num_periods, ar,
                intercept=intercept, log=log,
            )
            for initial_data in iter_initial_data
        ))
        self.set_data(span, new_data, )

    #]


for n in ("extrapolate", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )


def _extrapolate(
    initial: _np.ndarray,
    num_periods: int,
    ar: Iterable[Real],
    intercept: Real = 0,
    log: bool = False,
) -> None:
    """
    """
    #[
    initial = _np.flip(initial.flatten(), )
    if log:
        initial = _np.log(initial)
    intercept = _np.full((num_periods, ), intercept, dtype=_np.float64, )
    ma = (1, )
    zi = _sp.signal.lfiltic(ma, ar, initial, )
    out, *_ = _sp.signal.lfilter(ma, ar, intercept, zi=zi, axis=0, )
    if log:
        out = _np.exp(out)
    return out.reshape(-1, 1, )
    #]



"""
"""


#[
from __future__ import annotations

from typing import (Any, Protocol, Callable, )
from collections.abc import (Iterable, )
import numpy as _np
import scipy as _sp

from .. import equations as _equations
from .. import quantities as _quantities
from ..incidences import main as _incidences
from ..aldi import differentiators as _differentiators
from ..aldi import maps as _maps
#]


class Jacobian:
    """
    """
    #[

    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        wrt_something: Iterable[Any],
        qid_to_logly: dict[int, bool],
        /,
        *,
        context: dict[str, Callable] | None = None,
        first_column_to_eval: int | None = None,
        num_columns_to_eval: int = 1,
        sparse_jacobian: bool = False,
    ) -> None:
        """
        """
        #
        # Choose dense or sparse array creator
        self.is_sparse = sparse_jacobian
        self._create_jacobian = (
            _create_sparse_jacobian
            if self.is_sparse else _create_dense_jacobian
        )
        #
        # Extract eids from equations
        eids = tuple(eqn.id for eqn in equations)
        wrt_something = tuple(wrt_something)
        #
        # Collect w.r.t. items (tokens, qids) which each equation is to be
        # differentiated
        eid_to_wrts = self._create_eid_to_wrts(equations, wrt_something, )
        #
        # Collect all tokens in the euqations, and find the minimum shift
        all_tokens = set(_equations.generate_all_tokens_from_equations(equations, ), )
        min_shift = _incidences.get_min_shift(all_tokens, )
        self._first_column_to_eval = -min_shift
        #
        # Create the map from eids to rhs offsets; the offset is the number
        # of rows in the Jacobian matrix that precede the equation
        eid_to_rhs_offset = \
            _maps.create_eid_to_rhs_offset(eids, eid_to_wrts, )
        #
        self._shape = len(eids), len(wrt_something),
        #
        self._map = _maps.ArrayMap(
            eids,
            eid_to_wrts,
            wrt_something,
            eid_to_rhs_offset,
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        # self is the Atom factory
        atom_factory = self
        self._aldi_context = _differentiators.Context(
            atom_factory,
            equations,
            eid_to_wrts=eid_to_wrts,
            qid_to_logly=qid_to_logly,
            context=context,
        )

    def _create_eid_to_wrts(
        self,
        equations: Iterable[_equations.Equation],
        wrts: Iterable[Any],
        /,
    ) -> dict[int, tuple[Any, ...]]:
        """
        """
        ...

    #]


def _create_dense_jacobian(shape, diff_array, map, /, ) -> _np.ndarray:
    """
    Create Jacobian as numpy array
    """
    #[
    J = _np.zeros(shape, dtype=float, )
    J[map.lhs] = diff_array[map.rhs]
    return J
    #]


def _create_sparse_jacobian(shape, diff_array, map, /, ) -> _sp.sparse.coo_matrix:
    """
    Create Jacobian as scipy sparse matrix
    """
    #[
    J = _sp.sparse.coo_matrix(
        (diff_array[map.rhs], (map.lhs[0], map.lhs[1], )),
        shape, dtype=float,
    )
    return J
    #]





# Write a module to read pyproject.toml and bump the version number in it,
# depending on the argument passed to the script.

from typing import (Literal, )
import argparse
import toml


def _upgrade_version_string(version: str, bump: Literal["major", "minor", "patch"]) -> str:
    """Bump the version number"""
    major_minor_patch = version.split(".")
    index = next(i for i, e in enumerate(("major", "minor", "patch", )) if e == bump)
    major_minor_patch[index] = str(int(major_minor_patch[index]) + 1)
    return ".".join(major_minor_patch, )


def main(args, ):
    with open(args.source_path, "rt", ) as f:
        file = toml.load(f, )
    current_version = file["project"]["version"]
    bumped_version = _upgrade_version_string(current_version, args.release_type, )
    file["project"]["version"] = bumped_version
    print(bumped_version, )
    if args.target_path:
        with open(args.target_path, "wt", ) as f:
            f.write(toml.dumps(file, ), )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--release_type", choices=["major", "minor", "patch"])
    parser.add_argument("--source_path", )
    parser.add_argument("--target_path", default=None, )
    args = parser.parse_args()
    main(args, )

